"""Module with many classes that are further extended
in the other physbiblio.gui modules.

This file is part of the physbiblio package.
"""
import time
import traceback
from queue import Empty
from weakref import WeakValueDictionary

import PySide6
from PySide6.QtCore import (
    QAbstractItemModel,
    QAbstractTableModel,
    QModelIndex,
    QObject,
    QSortFilterProxyModel,
    Qt,
    QThread,
    QUrl,
    Signal,
)
from PySide6.QtGui import QAction, QDesktopServices, QGuiApplication, QPainter, QPixmap
from PySide6.QtWidgets import (
    QAbstractItemView,
    QComboBox,
    QDialog,
    QGridLayout,
    QLabel,
    QLineEdit,
    QMenu,
    QStyle,
    QTableView,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
)
from shiboken6 import VoidPtr

try:
    import physbiblio.gui.resourcesPyside6
    from physbiblio.database import catString, pBDB
    from physbiblio.errors import PBErrorManagerClass, pBLogger
    from physbiblio.pdf import pBPDF
    from physbiblio.strings.gui import CommonClassesStrings as ccstr
    from physbiblio.view import ViewEntry
except ImportError:
    print("Could not find physbiblio and its modules!")
    print(traceback.format_exc())


def PtrKey(ptr):
    """Retrieve a key to identify the VoidPtr.
    This is needed because VoidPtr is not hashable
    """
    return str(ptr).split()[1]


class PBDialog(QDialog):
    """Extend QDialog with centerWindow"""

    def centerWindow(self):
        """Use the `QGuiApplication` to get the relevant information
        and center the dialog in the screen.
        """
        qr = self.frameGeometry()
        cp = QGuiApplication.primaryScreen().availableGeometry().center()
        qr.moveCenter(cp)
        self.move(qr.topLeft())

    def cleanLayout(self):
        """Delete the previous table widget and other layout items"""
        if self.layout() is None:
            return
        while True:
            o = self.layout().takeAt(0)
            if o is None:
                break
            o.widget().deleteLater()


class PBLabel(QLabel):
    """Extension of `QLabel` with text interaction flags
    which enable text selection with the mouse
    """

    def __init__(self, *args, **kwargs):
        """Extend `QLabel.__init__` and call `setTextInteractionFlags`
        to allow text selection with the mouse
        """
        QLabel.__init__(self, *args, **kwargs)
        self.setTextFormat(Qt.RichText)
        self.setOpenExternalLinks(True)
        self.setTextInteractionFlags(
            Qt.LinksAccessibleByMouse | Qt.TextSelectableByMouse
        )


class PBLabelRight(PBLabel):
    """Class that creates a right-aligned QLabel"""

    def __init__(self, label):
        """The constructor.

        Parameter:
            label: the text label to be passed to QLabel
        """
        super(PBLabelRight, self).__init__(label)
        self.setAlignment(Qt.AlignRight | Qt.AlignVCenter)


class PBLabelCenter(PBLabel):
    """Class that creates a center-aligned QLabel"""

    def __init__(self, label):
        """The constructor.

        Parameter:
            label: the text label to be passed to QLabel
        """
        super(PBLabelCenter, self).__init__(label)
        self.setAlignment(Qt.AlignCenter | Qt.AlignVCenter)


class PBComboBox(QComboBox):
    """Personalize QComboBox for faster construction"""

    def __init__(self, parent, fields, current=None):
        """Constructor.

        Parameters:
            parent: the parent widget
            fields: the list of (string or printable) contents
                to be added to the QComboBox
            current (default None): the value to be set as the initial value
        """
        super(PBComboBox, self).__init__(parent)
        for f in fields:
            self.addItem("%s" % f)
        if current is not None:
            try:
                self.setCurrentIndex(fields.index(current))
            except ValueError:
                pass


class PBAndOrCombo(PBComboBox):
    """Shortcut for generating a QComboBox with and/or"""

    def __init__(self, parent, current=None):
        """Constructor.

        Parameters:
            parent: the parent widget
            current (default None): the value to be set
                as selected at the beginning
        """
        super(PBAndOrCombo, self).__init__(
            parent, [ccstr.andC, ccstr.orC], current=current
        )


class PBTrueFalseCombo(PBComboBox):
    """Shortcut for generating a QComboBox with true/false"""

    def __init__(self, parent, current=None):
        """Constructor.

        Parameters:
            parent: the parent widget
            current (default None): the value to be set as selected
                at the beginning
        """
        super(PBTrueFalseCombo, self).__init__(
            parent, [ccstr.true, ccstr.false], current=current
        )


class ObjListWindow(PBDialog):
    """Create a window managing a list (of bibtexs or of experiments)"""

    def __init__(self, parent=None, gridLayout=False):
        """Init using parent class and create common definitions

        Parameters:
            parent: the parent object
            gridLayout (boolean, default False):
                if True, use a QGridLayout, otherwise a QVBoxLayout
        """
        super(ObjListWindow, self).__init__(parent)
        self.tableWidth = None
        self.proxyModel = None
        self.gridLayout = gridLayout
        self.filterInput = None
        self.proxyModel = None
        self.tableview = None
        if gridLayout:
            self.currLayout = QGridLayout(self)
        else:
            self.currLayout = QVBoxLayout(self)

    def triggeredContextMenuEvent(self, row, col, event):
        """Not implemented: requires a subclass"""
        raise NotImplementedError()

    def handleItemEntered(self, index):
        """Not implemented: requires a subclass"""
        raise NotImplementedError()

    def cellClick(self, index):
        """Not implemented: requires a subclass"""
        raise NotImplementedError()

    def cellDoubleClick(self, index):
        """Not implemented: requires a subclass"""
        raise NotImplementedError()

    def createTable(self, *args, **kwargs):
        """Not implemented: requires a subclass"""
        raise NotImplementedError()

    def changeFilter(self, string):
        """Change the filter of the current view.

        Parameter:
            string: the filter string to be matched
        """
        self.proxyModel.setFilterRegularExpression(str(string))

    def addFilterInput(self, placeholderText, gridPos=(1, 0)):
        """Add a `QLineEdit` to change the filter of the list.

        Parameter:
            placeholderText: the text to be shown
                when no filter is present
            gridPos (tuple): if gridLayout is active,
                the position of the `QLineEdit` in the `QGridLayout`
        """
        self.filterInput = QLineEdit("", self)
        self.filterInput.setPlaceholderText(placeholderText)
        self.filterInput.textChanged.connect(self.changeFilter)

        if self.gridLayout:
            self.currLayout.addWidget(self.filterInput, *gridPos)
        else:
            self.currLayout.addWidget(self.filterInput)
        self.filterInput.setFocus()

    def setProxyStuff(self, sortColumn, sortOrder):
        """Prepare the proxy model to filter and sort the view.

        Parameter:
            sortColumn: the index of the column to use
                for sorting at the beginning
            sortOrder: the order for sorting
                (`Qt.AscendingOrder` or `Qt.DescendingOrder`)
        """
        self.proxyModel = QSortFilterProxyModel(self)
        self.proxyModel.setSourceModel(self.tableModel)
        self.proxyModel.setFilterCaseSensitivity(Qt.CaseInsensitive)
        self.proxyModel.setSortCaseSensitivity(Qt.CaseInsensitive)
        self.proxyModel.setFilterKeyColumn(-1)

        self.tableview = PBTableView(self)
        self.tableview.setModel(self.proxyModel)
        self.tableview.setSortingEnabled(True)
        self.tableview.setMouseTracking(True)
        self.tableview.setSelectionBehavior(QAbstractItemView.SelectRows)
        try:
            self.tableview.sortByColumn(
                self.tableModel.header.index("bibkey"), Qt.AscendingOrder
            )
        except (IndexError, ValueError):
            pass
        self.tableview.sortByColumn(sortColumn, sortOrder)
        try:
            self.proxyModel.sort(
                self.tableModel.header.index("bibkey"), Qt.AscendingOrder
            )
        except (IndexError, ValueError):
            pass
        self.proxyModel.sort(sortColumn, sortOrder)
        self.currLayout.addWidget(self.tableview)

    def finalizeTable(self, gridPos=(1, 0)):
        """Resize the table to fit the contents,
        connect functions, add to layout

        Parameter:
            gridPos (tuple): if gridLayout is active,
            the position of the `QLineEdit` in the `QGridLayout`
        """
        self.tableview.resizeColumnsToContents()

        maxh = QGuiApplication.primaryScreen().availableGeometry().height()
        maxw = QGuiApplication.primaryScreen().availableGeometry().width()
        self.setMaximumHeight(maxh)
        self.setMaximumWidth(maxw)

        hwidth = self.tableview.horizontalHeader().length()
        swidth = self.tableview.style().pixelMetric(QStyle.PM_ScrollBarExtent)
        fwidth = self.tableview.frameWidth() * 2

        if self.tableWidth is None:
            if hwidth > maxw - (swidth + fwidth):
                self.tableWidth = maxw - (swidth + fwidth)
            else:
                self.tableWidth = hwidth + swidth + fwidth
        self.tableview.setFixedWidth(self.tableWidth)

        self.setMinimumHeight(600)

        self.tableview.resizeColumnsToContents()
        self.tableview.resizeRowsToContents()

        self.tableview.entered.connect(self.handleItemEntered)
        self.tableview.clicked.connect(self.cellClick)
        self.tableview.doubleClicked.connect(self.cellDoubleClick)

        if self.gridLayout:
            self.currLayout.addWidget(self.tableview, *gridPos)
        else:
            self.currLayout.addWidget(self.tableview)

    def recreateTable(self):
        """Delete the previous table widget and other layout items,
        then create new ones
        """
        self.cleanLayout()
        self.createTable()


class EditObjectWindow(PBDialog):
    """Create a window for editing or creating an experiment"""

    def __init__(self, parent=None):
        """Constructor.

        Parameter:
            parent: the parent object
        """
        super(EditObjectWindow, self).__init__(parent)
        self.textValues = {}
        self.result = False
        self.currGrid = None
        self.initUI()

    def keyPressEvent(self, e):
        """Intercept press keys and exit if escape is pressed

        Parameters:
            e: the `PySide6.QtGui.QKeyEvent`
        """
        if e.key() == Qt.Key_Escape:
            self.onCancel()

    def onCancel(self):
        """Set that the result should not be considered and exit"""
        self.result = False
        self.close()

    def onOk(self):
        """Set that the result should be considered and exit"""
        self.result = True
        self.close()

    def initUI(self):
        """Instantiate the `QGridLayout`"""
        self.currGrid = QGridLayout()
        self.currGrid.setSpacing(1)
        self.setLayout(self.currGrid)


class PBThread(QThread):
    """Extend `QThread`, but further extension is needed"""

    finished = Signal()

    def __init__(self, parent=None):
        """Construct the class using `QThread.__init__`

        Parameters:
            parent: the parent object (default None)
        """
        QThread.__init__(self, parent)

    def run(self):
        """Not implemented: requires a subclass"""
        raise NotImplementedError()

    def setStopFlag(self):
        """Not implemented: requires a subclass"""
        raise NotImplementedError()

    def start(self, *args, **kwargs):
        """Wait 0.3s and then call `QThread.start`
        (pass all the arguments)
        """
        time.sleep(0.3)
        QThread.start(self, *args, **kwargs)


class WriteStream(PBThread):
    """Class used to redirect prints to a window"""

    newText = Signal(str)

    finished = Signal()

    def __init__(self, queue, parent=None, *args, **kwargs):
        """Constructor

        Parameters:
            queue: a Queue instance
            parent (optional): the parent widget
        """
        super(WriteStream, self).__init__(parent, *args, **kwargs)
        self.queue = queue
        self.running = True

    def write(self, text):
        """Write the given text

        Parameters:
            text: the text to send to the output stream
        """
        self.queue.put(text)

    def run(self):
        """Run the thread"""
        last = time.time()
        text = ""
        while self.running:
            try:
                text += self.queue.get(timeout=0.2)
            except Empty:
                pass
            if text != "" and (len(text) > 1000 or time.time() - last > 0.1):
                self.newText.emit(text)
                last = time.time()
                text = ""
        if text != "":
            self.newText.emit(text)
        self.finished.emit()


class PBTableView(QTableView):
    """Extension of `QTableView`, used to define the contextMenuEvent"""

    def contextMenuEvent(self, event):
        """Connect the context menu event to the parent function

        Parameter:
            event: the `PySide6.QtGui.QContextMenuEvent`
        """
        self.parent().triggeredContextMenuEvent(
            self.rowAt(event.y()), self.columnAt(event.x()), event
        )


class PBTableModel(QAbstractTableModel):
    """Extension of `QAbstractTableModel`,
    used for experiments and bibtex entries
    """

    def __init__(self, parent, header, ask=False, previous=[], *args):
        """Constructor, based on `QAbstractTableModel.__init__`

        Parameters:
            parent: the parent widget
            header: the list of column names which constitute the table
            ask (boolean, default False):
                when True, allow to select the lines with a checkbox
            previous: the list of lines which must be selected at the beginning
        """
        QAbstractTableModel.__init__(self, parent, *args)
        self.header = header
        self.parentObj = parent
        self.previous = previous
        self.ask = ask
        self.selectedElements = {}
        self.painter = None

    def parent(self):
        """Return the parent object"""
        return self.parentObj

    def changeAsk(self, new=None):
        """Change the value of `self.ask` to show/hide checkboxes

        Parameters:
            new: `None` to just swap the current value of `self.ask`,
                `True` or `False` to set it to the desired value
        """
        self.layoutAboutToBeChanged.emit()
        if new is None:
            self.ask = not self.ask
        elif new in (True, False):
            self.ask = new
        self.layoutChanged.emit()

    def getIdentifier(self, element):
        """Not implemented: requires a subclass"""
        raise NotImplementedError()

    def prepareSelected(self):
        """Fill the dictionary `self.selectedElements`
        according to previous selection
        """
        self.layoutAboutToBeChanged.emit()
        self.selectedElements = {}
        try:
            for bib in self.dataList:
                self.selectedElements[self.getIdentifier(bib)] = False
        except AttributeError:
            pBLogger.exception(ccstr.dataListNotDef)
        for prevK in self.previous:
            try:
                if self.selectedElements[prevK] == False:
                    self.selectedElements[prevK] = True
            except (KeyError, IndexError):
                pBLogger.exception(ccstr.invalidIdentif % (prevK))
        self.layoutChanged.emit()

    def selectAll(self):
        """Select all the available rows"""
        self.layoutAboutToBeChanged.emit()
        for key in self.selectedElements.keys():
            self.selectedElements[key] = True
        self.layoutChanged.emit()

    def unselectAll(self):
        """Unselect all the available rows"""
        self.layoutAboutToBeChanged.emit()
        for key in self.selectedElements.keys():
            self.selectedElements[key] = False
        self.layoutChanged.emit()

    def addImage(self, imagePath, height):
        """Create a cell containing an image

        Parameters:
            imagePath: the path of the image
            height: the height to give to the image

        Output:
            a QPixmap
        """
        return QPixmap(imagePath).scaledToHeight(height)

    def addImages(self, imagePaths, outHeight, height=48):
        """Create a cell containing multiple images, using a `QPainter`

        Parameters:
            imagePaths: a list of image paths
                for creating the single `QPixmap`s
            outHeight: the final height for rescaling the image
            height (default 48): the height and width
                to be used when painting

        Output:
            a `QPixmap`
        """
        width = len(imagePaths) * height
        pm = QPixmap(width, height)
        pm.fill(Qt.transparent)
        self.painter = QPainter(pm)
        for i, img in enumerate(imagePaths):
            self.painter.drawPixmap(i * height, 0, QPixmap(img))
        self.painter.end()
        return pm.scaledToHeight(outHeight)

    def rowCount(self, parent=None):
        """Count the rows of the given model based on the header

        Parameter:
            parent: `QModelIndex` (required by the parent signature)

        Output:
            the number od rows or zero (if error occurred)
        """
        try:
            return len(self.dataList)
        except (TypeError, AttributeError):
            return 0

    def columnCount(self, parent=None):
        """Count the columns of the given model based on the header

        Parameter:
            parent: `QModelIndex` (required by the parent signature)

        Output:
            the number of columns or zero (if error occurred)
        """
        try:
            return len(self.header)
        except TypeError:
            return 0

    def data(self, index, role):
        """Not implemented: requires a subclass"""
        raise NotImplementedError()

    def flags(self, index):
        """Determine the flags of a given item

        Parameter:
            index: a `QModelIndex`

        Output:
            None if the index is not valid
            if `self.ask` and first column, show checkboxes:
            Qt.ItemIsUserCheckable | Qt.ItemIsEditable |
                Qt.ItemIsEnabled | Qt.ItemIsSelectable
            all the other cases: Qt.ItemIsEnabled | Qt.ItemIsSelectable
        """
        if not index.isValid():
            return Qt.NoItemFlags
        if index.column() == 0 and self.ask:
            return (
                Qt.ItemIsUserCheckable
                | Qt.ItemIsEditable
                | Qt.ItemIsEnabled
                | Qt.ItemIsSelectable
            )
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def headerData(self, col, orientation, role):
        """Obtain column name if correctly asked

        Parameters:
            col: the column index in `self.header`
            orientation: int from `Qt.Orientation`
            role: int from `Qt.ItemDataRole`

        Output:
            the column name or `None`
        """
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self.header[col]
        return None

    def setData(self, index, value, role):
        """Not implemented: requires a subclass"""
        raise NotImplementedError()


# https://www.hardcoded.net/articles/using_qtreeview_with_qabstractitemmodel
class TreeNode(QObject):
    """Create an object that will work as a tree node"""

    # map `void*` -> `TreeNode` instance
    # see https://bugreports.qt.io/browse/PYSIDE-883
    _instances = WeakValueDictionary()

    @classmethod
    def cast(cls, obj):
        """Convert a VoidPtr to TreeNode, if possible and if necessary.

        Parameters:
            cls: the class instance
            obj: an index, from which the VoidPtr is extracted
                by internalPointer

        Output:
            should be a TreeNode instance
        """
        return obj.internalPointer()

    def __init__(self, parent, row):
        """Constructor, set basic properties

        Parameters:
            parent: the parent node
            row: the content of the data row
        """
        super(TreeNode, self).__init__()
        self.parentObj = parent
        self.row = row
        self.subnodes = self._getChildren()

    def parent(self):
        """Return the parent object"""
        return self.parentObj

    def _getChildren(self):
        """Not implemented: requires a subclass"""
        raise NotImplementedError()


class TreeModel(QAbstractItemModel):
    """Model for a tree structure."""

    def __init__(self):
        """Class constructor. Calls `_getRootNodes`
        to build the tree structure
        """
        QAbstractItemModel.__init__(self)
        self.rootNodes = self._getRootNodes()

    def _getRootNodes(self):
        """Not implemented: requires a subclass"""
        raise NotImplementedError()

    def index(self, row, column, parent=QModelIndex()):
        """Retrieve the `QModelIndex` of the requested object

        Parameters:
            row (int): the row index
            column (int): the column index
            parent: the parent `QModelIndex`

        Output:
            A `QModelIndex` instance
        """
        if not isinstance(parent, QModelIndex):
            pBLogger.debug(ccstr.invalidParentTM % (parent, "index"), exc_info=True)
            return QModelIndex()
        if not parent.isValid():
            try:
                return self.createIndex(row, column, self.rootNodes[row])
            except IndexError:
                return QModelIndex()
        parentNode = TreeNode.cast(parent)
        if parentNode is None:
            return QModelIndex()
        try:
            return self.createIndex(row, column, parentNode.subnodes[row])
        except IndexError:
            return QModelIndex()

    def parent(self, index):
        """Retrieve the `QModelIndex` of the parent of
        the item with the given index, if it exists,
        or an invalid `QModelIndex` instead

        Parameters:
            index: the `QModelIndex` of the node

        Output:
            A `QModelIndex` instance
        """
        if not isinstance(index, QModelIndex):
            pBLogger.debug(ccstr.invalidIndexTMP % index, exc_info=True)
            return QModelIndex()
        if not index.isValid():
            return QModelIndex()
        node = TreeNode.cast(index)
        if node is None:
            return QModelIndex()
        nodeParent = node.parent()
        if nodeParent is None:
            return QModelIndex()
        return self.createIndex(nodeParent.row, 0, nodeParent)

    def rowCount(self, parent=QModelIndex()):
        """Count the rows in a given tree branch

        Parameter:
            parent: the `QModelIndex` of the branch parent

        Output:
            the line number
        """
        if not isinstance(parent, QModelIndex):
            pBLogger.debug(ccstr.invalidParentTM % (parent, "rowCount"), exc_info=True)
            return len(self.rootNodes)
        if not parent.isValid():
            return len(self.rootNodes)
        node = TreeNode.cast(parent)
        if node is None:
            return len(self.rootNodes)
        return len(node.subnodes)


class NamedElement(object):
    """Basic object for the tree structure of categories"""

    def __init__(self, idCat, name, subelements):
        """Class constructor, set some element properties

        Parameters:
            idCat: the category id
            name: the name of the category
            subelements: the list of children elements
        """
        self.idCat = idCat
        self.name = name
        self.text = catString(idCat, pBDB)
        self.subelements = subelements


class NamedNode(TreeNode):
    """Extend `TreeNode` to work with `NamedElement`"""

    def __init__(self, element, parent, row):
        """Define `self.element` and call `TreeNode.__init__`

        Parameters:
            element: the `NamedElement` of the object
            parent: the parent node
            row: the row index
        """
        self.element = element
        TreeNode.__init__(self, parent, row)

    def _getChildren(self):
        """Return a list of `NamedNode`s, containing the children nodes.
        Overrides `TreeNode._getChildren`
        """
        return [
            NamedNode(elem, self, index)
            for index, elem in enumerate(self.element.subelements)
        ]


# http://gaganpreet.in/blog/2013/07/04/qtreeview-and-custom-filter-models/
class LeafFilterProxyModel(QSortFilterProxyModel):
    """Class to override the following behaviour:
        If a parent item doesn't match the filter,
        none of its children will be shown.

    This Model matches items which are descendants
    or ascendants of matching items.
    """

    def filterAcceptsRow(self, row_num, source_parent):
        """Overriding the parent function

        Parameters:
            row_num: the row number
            source_parent: the parent node in the tree
        """
        # Check if the current row matches
        if self.filterAcceptsRowItself(row_num, source_parent):
            return True

        # Traverse up all the way to root and check if any of them match
        if self.filterAcceptsAnyParent(source_parent):
            return True

        # Finally, check if any of the children match
        return self.hasAcceptedChildren(row_num, source_parent)

    def filterAcceptsRowItself(self, row_num, parent):
        """New name for the original `filterAcceptsRow`,
        which has been overridden

        Parameters:
            row_num: the row number
            parent: the parent node in the tree
        """
        return super(LeafFilterProxyModel, self).filterAcceptsRow(row_num, parent)

    def filterAcceptsAnyParent(self, parent):
        """Traverse to the root node and check if any of the
        ancestors match the filter

        Parameter:
            parent: the parent node in the tree
        """
        while parent.isValid():
            if self.filterAcceptsRowItself(parent.row(), parent.parent()):
                return True
            parent = parent.parent()
        return False

    def hasAcceptedChildren(self, row_num, parent):
        """Starting from the current node as root, traverse all
        the descendants and test if any of the children match

        Parameters:
            row_num: the row number
            parent: the parent node in the tree
        """
        model = self.sourceModel()
        source_index = model.index(row_num, 0, parent)

        children_count = model.rowCount(source_index)
        for i in range(children_count):
            if self.filterAcceptsRow(i, source_index):
                return True
        return False


class PBDDTableWidget(QTableWidget):
    """Drag and drop extension of QTableWidget"""

    def __init__(self, parent, header):
        """Set some properties and settings.

        Parameters:
            header: the title of the column
        """
        super(PBDDTableWidget, self).__init__(parent)
        self.setColumnCount(1)
        self.setHorizontalHeaderLabels([header])
        self.setDragEnabled(True)
        self.setAcceptDrops(True)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setDragDropOverwriteMode(False)
        self.lastDropRow = None

    # Override this method to get the correct row index for insertion
    def dropMimeData(self, row, col, mimeData, action):
        """Overridden method to get the row index for insertion

        Parameters:
            row: the index of the dropped row
            col, mimeData, action: not used params
                (see the signature of `QTableWidget.dropMimeData`)
        """
        self.lastDropRow = row
        return True

    def dropEvent(self, event):
        """Accept dropEvent and move the selected item
        to the new position

        Parameter:
            event: a `QDropEvent`
        """
        # The QTableWidget from which selected rows will be moved
        sender = event.source()

        # Default dropEvent method fires dropMimeData
        # with appropriate parameters (we're interested in the row index).
        super(PBDDTableWidget, self).dropEvent(event)
        # Now we know where to insert selected row(s)
        dropRow = self.lastDropRow

        selectedRows = sender.getselectedRowsFast()

        # Allocate space for transfer
        for _ in selectedRows:
            self.insertRow(dropRow)

        # if sender == receiver (self),
        # after creating new empty rows selected rows might change
        # their locations
        sel_rows_offsets = [
            0 if self != sender or srow < dropRow else len(selectedRows)
            for srow in selectedRows
        ]
        selectedRows = [
            row + offset for row, offset in zip(selectedRows, sel_rows_offsets)
        ]

        # copy content of selected rows into empty ones
        for i, srow in enumerate(selectedRows):
            for j in range(self.columnCount()):
                item = sender.item(srow, j)
                if item:
                    source = QTableWidgetItem(item)
                    self.setItem(dropRow + i, j, source)

        # delete selected rows
        for srow in reversed(selectedRows):
            sender.removeRow(srow)

        event.accept()

    def getselectedRowsFast(self):
        """Return the list of selected rows

        Output:
            a list with the row indexes of the selected rows
        """
        selectedRows = []
        for item in self.selectedItems():
            row = item.row()
            text = item.text()
            if row not in selectedRows and text != "bibkey":
                selectedRows.append(row)
        return sorted(selectedRows)


class PBMenu(QMenu):
    """Extend `QMenu` for faster menu build"""

    def __init__(self, parent=None):
        """Construct the element defining some basic properties

        Parameter:
            parent: the parent widget
        """
        super(PBMenu, self).__init__(parent)
        self.possibleActions = []
        self.result = False

    def fillMenu(self):
        """Add actions, separators or submenus according to the content
        of `self.possibleActions` (recursively for submenus)
        """
        for act in self.possibleActions:
            if act is None:
                self.addSeparator()
            elif isinstance(act, list):
                submenu = PBMenu()
                submenu.setTitle(act[0])
                submenu.possibleActions = act[1]
                submenu.fillMenu()
                self.addMenu(submenu)
            elif isinstance(act, dict):
                submenu = PBMenu()
                submenu.setTitle(act["title"])
                submenu.possibleActions = act["actions"]
                submenu.fillMenu()
                try:
                    submenu.setToolTipsVisible(act["toolTipsVisible"])
                except KeyError:
                    pass
                self.addMenu(submenu)
            elif isinstance(act, QAction):
                self.addAction(act)

    def keyPressEvent(self, e):
        """Intercept press keys and exit if escape is pressed

        Parameters:
            e: the `PySide6.QtGui.QKeyEvent`
        """
        if e.key() == Qt.Key_Escape:
            self.close()


class GUIViewEntry(ViewEntry):
    """Extends the ViewEntry class to work with QtGui.QDesktopServices"""

    def openLink(self, key, arg="", fileArg=None):
        """Use `QDesktopServices` to open an url using
        the system default applications

        Parameters:
            key: the entry key or the link (if `arg` == "link")
            arg:
                if `arg` == "file", `fileArg` must be the file name
                if `arg` == "link", `key` must be the link to be opened
                for any other values, the link will be generated using
                    the `physbiblio.view.viewWntry.getLink` method
            fileArg: the file name if `arg` == "file", or
                the argument passed to `physbiblio.view.viewWntry.getLink`
                if needed
        """
        if isinstance(key, list):
            for k in key:
                self.openLink(k, arg, fileArg)
        else:
            if arg == "file":
                url = QUrl.fromLocalFile(fileArg)
            elif arg == "link":
                url = QUrl(key)
            else:
                link = self.getLink(key, arg=arg, fileArg=fileArg)
                url = QUrl(link)
            if QDesktopServices.openUrl(url):
                pBLogger.debug(ccstr.openSuccess % (url.toString(), key))
            else:
                pBLogger.warning(ccstr.openFailed % key)


pBGuiView = GUIViewEntry()


class PBImportedTableModel(PBTableModel):
    """Extend `PBTableModel` to manage the selection
    during the import of entries
    """

    def __init__(self, parent, bibdict, header, idName="ID", *args):
        """Set some properties and settings

        Parameters:
            parent: the parent widget (pass to `PBTableModel.__init__`)
            bibdict: a dictionary with the info
                of the imported bibtex entries.
                It should contain dictionaries with two items:
                    "bibpars", "exist"
            header: the header names to be passed
                to `PBTableModel.__init__`
            idName: the key of the field representing
                the unique ID of the item
        """
        self.typeClass = "imports"
        self.idName = idName
        self.bibsOrder = [k for k in sorted(bibdict.keys())]
        self.dataList = [bibdict[k]["bibpars"] for k in self.bibsOrder]
        self.existList = [bibdict[k]["exist"] for k in self.bibsOrder]
        PBTableModel.__init__(self, parent, header, *args)
        self.prepareSelected()

    def getIdentifier(self, element):
        """Return the unique identifier of the given element

        Parameters:
            element: a dictionary
        """
        return element[self.idName]

    def data(self, index, role):
        """Return the data for the requested cell and role

        Parameters:
            index: a `QModelIndex`
            role: the requested role for the given cell
        """
        if not index.isValid():
            return None
        row = index.row()
        column = index.column()
        try:
            value = self.dataList[row][self.header[column]]
        except (IndexError, KeyError):
            pBLogger.debug(ccstr.missElement, exc_info=True)
            return None

        if role == Qt.CheckStateRole and column == 0 and not self.existList[row]:
            if self.selectedElements[self.dataList[row][self.idName]] == False:
                return Qt.Unchecked
            else:
                return Qt.Checked
        if (
            role in (Qt.EditRole, Qt.DisplayRole)
            and column == 0
            and self.existList[row]
        ):
            return value + ccstr.alreadyExisting
        if role == Qt.EditRole:
            return value
        if role == Qt.DisplayRole:
            return value
        return None

    def setData(self, index, value, role):
        """Set the selection data for a given index

        Parameters:
            index: a `QModelIndex`
            value: `Qt.Checked` to set the row as selected
            role: `Qt.CheckStateRole` for performing an action,
                or any other role
        """
        if not index.isValid():
            return False
        if role == Qt.CheckStateRole and index.column() == 0:
            if Qt.CheckState(value) == Qt.Checked:
                self.selectedElements[self.dataList[index.row()][self.idName]] = True
            else:
                self.selectedElements[self.dataList[index.row()][self.idName]] = False
        self.dataChanged.emit(index, index)
        return True

    def flags(self, index):
        """Return the flags for the requested row

        Parameters:
            index: a `QModelIndex`
        """
        if not index.isValid():
            return Qt.NoItemFlags
        if index.column() == 0 and not self.existList[index.row()]:
            return (
                Qt.ItemIsUserCheckable
                | Qt.ItemIsEditable
                | Qt.ItemIsEnabled
                | Qt.ItemIsSelectable
            )
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable


class ObjectWithSignal(QObject):
    """Being a QObject is necessary to create a signal and use it"""

    customSignal = Signal()



#!/usr/bin/env python
"""Test file for the physbiblio.export module.

This file is part of the physbiblio package.
"""
import os
import traceback
import unittest
from unittest.mock import patch

import bibtexparser

try:
    from physbiblio.config import pbConfig
    from physbiblio.database import pBDB
    from physbiblio.export import pBExport
    from physbiblio.setuptests import *
except ImportError:
    print("Could not find physbiblio and its modules!")
    raise
except Exception:
    print(traceback.format_exc())


@unittest.skipIf(skipTestsSettings.long, "Long tests")
@patch("logging.Logger.debug")
@patch("logging.Logger.info")
@patch("logging.Logger.warning")
@patch("logging.Logger.exception")
class TestExportMethods(unittest.TestCase):
    """Tests for methods in physbiblio.export"""

    def setUp(self):
        """Generate filenames"""
        self.testBibName = os.path.join(pbConfig.dataPath, "tests_%s.bib" % today_ymd)
        self.testTexName = os.path.join(pbConfig.dataPath, "tests_%s.tex" % today_ymd)

    def tearDown(self):
        """remove files"""
        if os.path.exists(self.testBibName):
            os.remove(self.testBibName)
        if os.path.exists(self.testTexName):
            os.remove(self.testTexName)

    def test_backup(self, *args):
        """Test backup file creation and related"""
        emptyFileName = self.testBibName
        if os.path.exists(emptyFileName):
            os.remove(emptyFileName)
        if os.path.exists(emptyFileName + pBExport.backupExtension):
            os.remove(emptyFileName + pBExport.backupExtension)
        self.assertFalse(pBExport.backupCopy(emptyFileName))
        self.assertFalse(os.path.exists(emptyFileName))
        self.assertFalse(os.path.exists(emptyFileName + pBExport.backupExtension))
        open(emptyFileName, "a").close()
        self.assertTrue(pBExport.backupCopy(emptyFileName))
        self.assertTrue(os.path.exists(emptyFileName + pBExport.backupExtension))
        os.remove(emptyFileName)
        self.assertFalse(os.path.exists(emptyFileName))
        self.assertTrue(pBExport.restoreBackupCopy(emptyFileName))
        self.assertTrue(os.path.exists(emptyFileName))
        with patch("os.path.isfile", side_effect=[False, True, True]) as _isf, patch(
            "shutil.copy2", side_effect=[True, IOError]
        ) as _cp2, patch("logging.Logger.exception") as _ex:
            self.assertFalse(pBExport.restoreBackupCopy(emptyFileName))
            self.assertTrue(pBExport.restoreBackupCopy(emptyFileName))
            self.assertFalse(pBExport.restoreBackupCopy(emptyFileName))
            _ex.assert_called_once()
        self.assertTrue(pBExport.rmBackupCopy(emptyFileName))
        self.assertFalse(os.path.exists(emptyFileName + pBExport.backupExtension))

    def test_offlineExports(self, *args):
        """Test of offline export functions exportSelected,
        updateExportedBib
        """
        testBibName = self.testBibName
        sampleList = [
            {"bibtex": '@Article{empty,\nauthor="me",\ntitle="no"\n}'},
            {"bibtex": '@Article{empty2,\nauthor="me2",\ntitle="yes"\n}'},
        ]
        sampleTxt = (
            '@Article{empty,\nauthor="me",\ntitle="no"\n}\n'
            + '@Article{empty2,\nauthor="me2",\ntitle="yes"\n}\n'
        )
        pBDB.bibs.lastFetched = sampleList
        pBExport.exportLast(testBibName)
        with open(testBibName) as f:
            self.assertEqual(f.read(), sampleTxt)
        os.remove(testBibName)

        sampleList = [{"bibtex": '@Article{empty2,\nauthor="me2",\ntitle="{yes}",\n}'}]
        sampleTxt = '@Article{empty2,\nauthor="me2",\ntitle="{yes}",\n}\n'
        pBExport.exportSelected(testBibName, sampleList)
        with open(testBibName) as f:
            self.assertEqual(f.read(), sampleTxt)

        with patch(
            "physbiblio.database.Entries.getByBibkey",
            return_value=[
                {
                    "bibtexDict": {
                        "ID": "empty2",
                        "ENTRYTYPE": "Article",
                        "author": "me et al",
                        "title": "yes",
                    }
                }
            ],
            autospec=True,
        ) as _mock:
            pBExport.updateExportedBib(testBibName, overwrite=True)
        with open(testBibName) as f:
            self.assertEqual(
                f.read().replace(" ", "").replace("\n", ""),
                sampleTxt.replace("me2", "me et al").replace(" ", "").replace("\n", ""),
            )
        pBExport.rmBackupCopy(testBibName)

    def test_exportAll(self, *args):
        """Test of exportAll"""
        testBibName = self.testBibName
        sampleList = [
            {"bibtex": '@Article{empty,\nauthor="me",\ntitle="no"\n}'},
            {"bibtex": '@Article{empty2,\nauthor="me2",\ntitle="yes"\n}'},
        ]
        sampleTxt = (
            '@Article{empty,\nauthor="me",\ntitle="no"\n}\n'
            + '@Article{empty2,\nauthor="me2",\ntitle="yes"\n}\n'
        )
        with patch(
            "physbiblio.database.Entries.fetchCursor",
            return_value=sampleList,
            autospec=True,
        ) as _curs:
            pBExport.exportAll(testBibName)
        with open(testBibName) as f:
            self.assertEqual(f.read(), sampleTxt)

    def test_exportForTexFile(self, *args):
        """test exportForTexFile function with a fake tex and database"""
        testBibName = self.testBibName
        testTexName = self.testTexName
        texString = (
            "\cite{empty,prova., empty2+}\citep{empty2+}"
            + "\citet{Gariazzo:2015rra}, \citet{Gariazzo:2017rra}\n"
        )
        with open(testTexName, "w") as f:
            f.write(texString)
        with open(testTexName) as f:
            self.assertEqual(f.read(), texString)
        sampleList = [
            {
                "bibkey": "empty",
                "bibtex": '@Article{empty,\n        author = "me",\n         '
                + 'title = "{no}",\n}',
            },
            {
                "bibkey": "empty2+",
                "bibtex": '@Article{empty2+,\nauthor="me2",\ntitle="yes"\n}',
            },
        ]

        self.assertFalse(os.path.exists(testBibName))
        with patch(
            "physbiblio.database.CatsEntries.insert", return_value=True, autospec=True
        ) as _ceins, patch(
            "physbiblio.database.Entries.getByBibtex",
            side_effect=[
                [
                    {
                        "bibkey": "empty",
                        "bibtexDict": {},
                        "bibtex": '@Article{empty,\nauthor="me",\ntitle="no"\n}',
                    }
                ],
                [],
                [
                    {
                        "bibkey": "empty2+",
                        "bibtexDict": {},
                        "bibtex": '@Article{empty2+,\nauthor="me2",\ntitle="yes"\n}',
                    }
                ],
                [
                    {
                        "bibkey": "Gariazzo:2015rra",
                        "bibtexDict": {},
                        "bibtex": "@article{Gariazzo:2015rra,\nauthor= "
                        + '"Gariazzo, S. and others",\ntitle='
                        + '"{Light sterile neutrinos}",\n}',
                    }
                ],
                [],
            ],
            autospec=True,
        ) as _getbbibt, patch(
            "physbiblio.database.Entries.getByKey",
            side_effect=[
                [
                    {
                        "bibkey": "Gariazzo:2015rra",
                        "bibtexDict": {},
                        "bibtex": "@article{Gariazzo:2015rra,\nauthor= "
                        + '"Gariazzo, S. and others",\ntitle='
                        + '"{Light sterile neutrinos}",\n}',
                    }
                ],
                [],
                [],
            ],
            autospec=True,
        ) as _getbbik, patch(
            "physbiblio.database.Entries.loadAndInsert",
            side_effect=[
                "@article{Gariazzo:2015rra,\nauthor= "
                + '"Gariazzo, S. and others",\ntitle='
                + '"{Light sterile neutrinos}",\n}',
                "@article{Gariazzo:2015rra,\nauthor= "
                + '"Gariazzo, S. and others",\ntitle='
                + '"{Light sterile neutrinos}",\n}',
                "",
            ],
            autospec=True,
        ) as _mock:
            output = pBExport.exportForTexFile(
                testTexName, testBibName, overwrite=True, autosave=False
            )
        self.assertEqual(
            output[0],
            ["empty", "prova.", "empty2+", "Gariazzo:2015rra", "Gariazzo:2017rra"],
        )  # requiredBibkeys
        self.assertEqual(output[1], ["prova.", "Gariazzo:2017rra"])  # missing
        self.assertEqual(output[2], ["Gariazzo:2015rra"])  # retrieved
        self.assertEqual(output[3], ["Gariazzo:2017rra"])  # notFound
        self.assertEqual(output[4], [])  # unexpected
        self.assertEqual(output[5], {"prova.": "Gariazzo:2015rra"})  # newKeys
        self.assertEqual(output[6], 2)  # warnings
        self.assertEqual(output[7], 5)  # total
        self.assertTrue(os.path.exists(testBibName))
        with open(testBibName) as f:
            newTextBib = f.read()
        for t in (
            "%file written by PhysBiblio\n",
            '@Article{empty,\n        author = "me",\n         title = '
            + '"{no}",\n}\n\n',
            '@Article{prova.,\n        author = "Gariazzo, S. and others",'
            + '\n         title = "{Light sterile neutrinos}",'
            + "\n}\n\n",
            '@Article{empty2+,\n        author = "me2",\n         title = '
            + '"{yes}",\n}\n\n',
            "@Article{Gariazzo:2015rra,\n        author = "
            + '"Gariazzo, S. and others",\n         title = '
            + '"{Light sterile neutrinos}",\n}\n\n',
        ):
            self.assertIn(t, newTextBib)

        with open(testTexName, "a") as f:
            f.write("\cite{newcite}")
        with patch(
            "physbiblio.database.CatsEntries.insert", return_value=True, autospec=True
        ) as _ceins, patch(
            "physbiblio.database.Entries.getByBibtex",
            side_effect=[
                [],
                [],
                [],
                [
                    {
                        "bibkey": "newcite",
                        "bibtexDict": {},
                        "bibtex": '@article{newcite,\nauthor= "myself",\ntitle='
                        + '"{some paper}",\n}',
                    }
                ],
            ],
            autospec=True,
        ) as _getbbibt, patch(
            "physbiblio.database.Entries.getByKey",
            side_effect=[
                [],
                [
                    {
                        "bibkey": "newcite",
                        "bibtexDict": {},
                        "bibtex": '@article{newcite,\nauthor= "myself",\ntitle='
                        + '"{some paper}",\n}',
                    }
                ],
            ],
            autospec=True,
        ) as _getbbik, patch(
            "physbiblio.database.Entries.loadAndInsert",
            side_effect=[
                "",
                '@article{newcite,\nauthor= "myself",\ntitle=' + '"{some paper}",\n}',
            ],
            autospec=True,
        ) as _mock:
            output = pBExport.exportForTexFile(testTexName, testBibName, autosave=False)
        self.assertEqual(output[0], ["Gariazzo:2017rra", "newcite"])  # requiredBibkeys
        self.assertEqual(output[1], ["Gariazzo:2017rra", "newcite"])  # missing
        self.assertEqual(output[2], ["newcite"])  # retrieved
        self.assertEqual(output[3], ["Gariazzo:2017rra"])  # notFound
        self.assertEqual(output[4], [])  # unexpected
        self.assertEqual(output[5], {})  # newKeys
        self.assertEqual(output[6], 1)  # warnings
        self.assertEqual(output[7], 6)  # total
        self.assertTrue(os.path.exists(testBibName))
        with open(testBibName) as f:
            newTextBib = f.read()
        for t in (
            "%file written by PhysBiblio\n",
            '@Article{empty,\n        author = "me",\n         title = '
            + '"{no}",\n}\n\n',
            '@Article{prova.,\n        author = "Gariazzo, S. and others"'
            + ',\n         title = "{Light sterile neutrinos}",\n}\n\n',
            '@Article{empty2+,\n        author = "me2",\n         title = '
            + '"{yes}",\n}\n\n',
            "@Article{Gariazzo:2015rra,\n        author = "
            + '"Gariazzo, S. and others",\n         title = '
            + '"{Light sterile neutrinos}",\n}\n\n',
            '@Article{newcite,\n        author = "myself",'
            + '\n         title = "{some paper}",\n}\n\n',
        ):
            self.assertIn(t, newTextBib)

        with open(testTexName, "w") as f:
            f.write("\cite{newcite:NOW18}")
        with patch(
            "physbiblio.database.Entries.getByBibtex",
            side_effect=[
                [
                    {
                        "bibkey": "newcite:now18",
                        "bibtexDict": {},
                        "bibtex": '@article{newcite:now18,\nauthor= "myself",\ntitle='
                        + '"{some paper}",\n}',
                    }
                ]
            ],
            autospec=True,
        ) as _getbbibt, patch(
            "physbiblio.database.Entries.loadAndInsert", return_value=[], autospec=True
        ) as _lai:
            output = pBExport.exportForTexFile(
                testTexName, testBibName, overwrite=True, autosave=False
            )
        self.assertEqual(output[0], ["newcite:NOW18"])  # requiredBibkeys
        self.assertEqual(output[1], [])  # missing
        self.assertEqual(output[2], [])  # retrieved
        self.assertEqual(output[3], [])  # notFound
        self.assertEqual(output[4], [])  # unexpected
        self.assertEqual(output[5], {})  # newKeys
        self.assertEqual(output[6], 0)  # warnings
        self.assertEqual(output[7], 1)  # total
        self.assertTrue(os.path.exists(testBibName))
        with open(testBibName) as f:
            newTextBib = f.read()
        for t in (
            "%file written by PhysBiblio\n",
            '@Article{newcite:NOW18,\n        author = "myself",\n         '
            + 'title = "{some paper}",\n}\n\n',
        ):
            self.assertIn(t, newTextBib)

        with open(testTexName, "w") as f:
            f.write("\cite{newcite:NOW18}\cite{bib1}\cite{bib2}")
        bibtex1 = '@Article{newcite,\nauthor="S. Gariazzo",\ntitle=' + '"{newtitle}"\n}'
        bibtex2 = (
            '@article{bib1,\nauthor="SG",\ntitle=' + '"{Light sterile neutrinos}",\n}'
        )
        bibtex3 = '@Article{bib2,\nauthor="me",\ntitle="title"\n}'
        with patch(
            "physbiblio.database.Entries.getByBibtex",
            side_effect=[
                [
                    {
                        "bibkey": "newcite:now18",
                        "bibtexDict": bibtexparser.loads(bibtex1).entries[0],
                        "bibtex": bibtex1,
                    }
                ],
                [
                    {
                        "bibkey": "bib1",
                        "bibtexDict": bibtexparser.loads(bibtex2).entries[0],
                        "bibtex": bibtex2,
                    }
                ],
                [
                    {
                        "bibkey": "bib2",
                        "bibtexDict": bibtexparser.loads(bibtex3).entries[0],
                        "bibtex": bibtex3,
                    }
                ],
            ],
            autospec=True,
        ) as _getbbibt, patch(
            "physbiblio.database.Entries.loadAndInsert", return_value=[], autospec=True
        ) as _lai:
            output = pBExport.exportForTexFile(
                testTexName, testBibName, autosave=False, updateExisting=True
            )
        self.assertEqual(output[0], ["bib1", "bib2"])  # requiredBibkeys
        self.assertEqual(output[1], [])  # missing
        self.assertEqual(output[2], [])  # retrieved
        self.assertEqual(output[3], [])  # notFound
        self.assertEqual(output[4], [])  # unexpected
        self.assertEqual(output[5], {})  # newKeys
        self.assertEqual(output[6], 0)  # warnings
        self.assertEqual(output[7], 3)  # total
        self.assertTrue(os.path.exists(testBibName))
        self.maxDiff = None
        with open(testBibName) as f:
            newTextBib = f.read()
        for t in (
            "%file written by PhysBiblio\n",
            '@Article{newcite:NOW18,\n        author = "S. Gariazzo",'
            + '\n         title = "{newtitle}",\n}\n\n',
            '@Article{bib1,\n        author = "SG",\n         title = '
            + '"{Light sterile neutrinos}",\n}\n\n',
            '@Article{bib2,\n        author = "me",\n         title = '
            + '"{title}",\n}\n\n',
        ):
            self.assertIn(t, newTextBib)

        with open(testTexName, "w") as f:
            f.write("\cite{newcite:NOW18}\cite{bib1}")
        bibtex1 = (
            '@Article{newcite:now18,\nauthor="S. Gariazzo",\ntitle=' + '"{newtitle}"\n}'
        )
        bibtex2 = (
            '@article{bib1,\nauthor="SG",\ntitle=' + '"{Light sterile neutrinos}",\n}'
        )
        with patch(
            "physbiblio.database.Entries.getByBibtex",
            side_effect=[
                [
                    {
                        "bibkey": "newcite:now18",
                        "bibtexDict": bibtexparser.loads(bibtex1).entries[0],
                        "bibtex": bibtex1,
                    }
                ],
                [
                    {
                        "bibkey": "bib1",
                        "bibtexDict": bibtexparser.loads(bibtex2).entries[0],
                        "bibtex": bibtex2,
                    }
                ],
            ],
            autospec=True,
        ) as _getbbibt, patch(
            "physbiblio.database.Entries.loadAndInsert", return_value=[], autospec=True
        ) as _lai:
            output = pBExport.exportForTexFile(
                testTexName, testBibName, autosave=False, removeUnused=True
            )
        self.assertEqual(output[0], [])  # requiredBibkeys
        self.assertEqual(output[1], [])  # missing
        self.assertEqual(output[2], [])  # retrieved
        self.assertEqual(output[3], [])  # notFound
        self.assertEqual(output[4], [])  # unexpected
        self.assertEqual(output[5], {})  # newKeys
        self.assertEqual(output[6], 0)  # warnings
        self.assertEqual(output[7], 2)  # total
        self.assertTrue(os.path.exists(testBibName))
        self.maxDiff = None
        with open(testBibName) as f:
            newTextBib = f.read()
        for t in (
            "%file written by PhysBiblio\n",
            '@Article{newcite:NOW18,\n        author = "S. Gariazzo",'
            + '\n         title = "{newtitle}",\n}\n\n',
            '@Article{bib1,\n        author = "SG",\n         '
            + 'title = "{Light sterile neutrinos}",\n}\n\n',
        ):
            self.assertIn(t, newTextBib)

        with open(testTexName, "w") as f:
            f.write("\cite{newcite}")
        bibtex1 = '@Article{newcite,\nauthor="S. Gariazzo",' + '\ntitle="{title}"\n}'
        with patch(
            "physbiblio.database.Entries.getByBibtex",
            side_effect=[
                [
                    {
                        "bibkey": "newcite",
                        "bibtexDict": bibtexparser.loads(bibtex1).entries[0],
                        "bibtex": bibtex1,
                    }
                ]
            ],
            autospec=True,
        ) as _getbbibt, patch(
            "physbiblio.database.Entries.loadAndInsert", return_value=[], autospec=True
        ) as _lai:
            output = pBExport.exportForTexFile(
                testTexName, testBibName, autosave=False, overwrite=True
            )
        self.assertEqual(output[0], ["newcite"])  # requiredBibkeys
        self.assertEqual(output[1], [])  # missing
        self.assertEqual(output[2], [])  # retrieved
        self.assertEqual(output[3], [])  # notFound
        self.assertEqual(output[4], [])  # unexpected
        self.assertEqual(output[5], {})  # newKeys
        self.assertEqual(output[6], 0)  # warnings
        self.assertEqual(output[7], 1)  # total
        self.assertTrue(os.path.exists(testBibName))
        self.maxDiff = None
        with open(testBibName) as f:
            newTextBib = f.read()
        for t in (
            "%file written by PhysBiblio\n",
            '@Article{newcite,\n        author = "S. Gariazzo",'
            + '\n         title = "{title}",\n}\n\n',
        ):
            self.assertIn(t, newTextBib)

        with open(testBibName, "w") as f:
            f.write(
                "%file written by PhysBiblio\n"
                + '@Article{empty,\n    author = "me",\n         title = '
                + '"{no}",\n}\n\n'
                + '@Article{prova,\n  author = "Gariazzo, S. and others",'
                + '\n         title = "{Light sterile neutrinos}",'
                + "\n\n"
                + '@Article{empty2,\n   author = "",\n         title = '
                + '"{yes}",\n}\n\n'
                + "@Article{Gariazzo:2015rra,\n        author = "
                + '"Gariazzo, S. and others",\n         %title = '
                + '"{Light sterile neutrinos}",\n}\n\n'
            )
        with patch(
            "physbiblio.database.Entries.getByBibtex", autospec=True
        ) as _getbbibt, patch(
            "physbiblio.database.Entries.loadAndInsert", return_value=[], autospec=True
        ) as _lai:
            output = pBExport.exportForTexFile(
                testTexName,
                testBibName,
                autosave=False,
                removeUnused=True,
                updateExisting=True,
            )
        self.assertEqual(output[0], ["newcite"])  # requiredBibkeys
        self.assertEqual(output[1], ["newcite"])  # missing
        self.assertEqual(output[2], [])  # retrieved
        self.assertEqual(output[3], ["newcite"])  # notFound
        self.assertEqual(output[4], [])  # unexpected
        self.assertEqual(output[5], {})  # newKeys
        self.assertEqual(output[6], 1)  # warnings
        self.assertEqual(output[7], 1)  # total
        self.assertTrue(os.path.exists(testBibName))
        with open(testBibName) as f:
            newTextBib = f.read()
        self.assertEqual(newTextBib, "%file written by PhysBiblio\n")

        os.remove(testBibName)
        with patch("logging.Logger.exception") as _er, patch(
            "physbiblio.database.Entries.loadAndInsert", return_value=[], autospec=True
        ) as _lai:
            output = pBExport.exportForTexFile(testTexName, testBibName, autosave=False)
            _er.assert_any_call("Cannot read file %s.\nCreating one." % testBibName)
        self.assertEqual(len(output), 8)
        self.assertTrue(os.path.exists(testBibName))
        with patch("logging.Logger.exception") as _ex, patch(
            "physbiblio.database.Entries.loadAndInsert", return_value=[], autospec=True
        ) as _lai:
            self.assertFalse(
                pBExport.exportForTexFile(
                    testTexName, "/surely/not/existing/path.bib", autosave=False
                )
            )
            _ex.assert_any_call("Cannot create file /surely/not/existing/path.bib!")

        texString = (
            "\cite{someA&A...123,prova., empty2+}\citep{empty2+}"
            + "\citet{Gariazzo:2015rra}, \citet{Gariazzo:2017rra}\n"
        )
        with open(testTexName, "w") as f:
            f.write(texString)
        with patch(
            "physbiblio.database.Entries.getByBibtex",
            return_value=[
                {
                    "bibkey": "newcite:now18",
                    "bibtexDict": bibtexparser.loads(bibtex1).entries[0],
                    "bibtex": bibtex1,
                }
            ],
            autospec=True,
        ) as _getbbibt, patch(
            "physbiblio.database.Entries.loadAndInsert", return_value=[], autospec=True
        ) as _lai:
            output = pBExport.exportForTexFile(testTexName, testBibName, autosave=False)
        self.assertEqual(
            output[0],
            [
                "someA&A...123",
                "prova.",
                "empty2+",
                "Gariazzo:2015rra",
                "Gariazzo:2017rra",
            ],
        )  # requiredBibkeys
        self.assertEqual(output[1], [])  # missing
        self.assertEqual(output[2], [])  # retrieved
        self.assertEqual(output[3], [])  # notFound
        self.assertEqual(output[4], [])  # unexpected
        self.assertEqual(output[5], {})  # newKeys
        self.assertEqual(output[6], 0)  # warnings
        self.assertEqual(output[7], 5)  # total

        with open(testBibName, "w") as f:
            f.write(
                "%file written by PhysBiblio\n"
                + '@Article{empty,\n    author = "me",\n         title = '
                + '"{no}",\n}\n\n'
                + '@Article{prova,\n  author = "Gariazzo, S. and others",'
                + '\n         title = "{Light sterile neutrinos}",'
                + "}\n\n"
                + '@Article{empty2,\n   author = "boh",\n         title = '
                + '"{yes}",\n}\n\n'
                + "@Article{Gariazzo:2015rra,\n        author = "
                + '"Gariazzo, S. and others",\n         title = '
                + '"{Light sterile neutrinos}",\n}\n\n'
            )
        with open(testTexName, "w") as f:
            f.write(
                "\cite{prova,empty2}\citep{empty}"
                + "\citet{Gariazzo:2015rra}, \citet{Gariazzo:2017rra}\n"
            )
        with patch(
            "physbiblio.database.Entries.getByBibtex", return_value=[], autospec=True
        ) as _getbbibt, patch(
            "physbiblio.database.Entries.loadAndInsert", return_value=[], autospec=True
        ) as _lai:
            output = pBExport.exportForTexFile(
                testTexName, testBibName, autosave=False, reorder=True
            )
        self.assertEqual(output[0], ["Gariazzo:2017rra"])  # requiredBibkeys
        self.assertEqual(output[1], ["Gariazzo:2017rra"])  # missing
        self.assertEqual(output[2], [])  # retrieved
        self.assertEqual(output[3], ["Gariazzo:2017rra"])  # notFound
        self.assertEqual(output[4], [])  # unexpected
        self.assertEqual(output[5], {})  # newKeys
        self.assertEqual(output[6], 1)  # warnings
        self.assertEqual(output[7], 5)  # total
        with open(testBibName) as f:
            newTextBib = f.readlines()
        useful = [l for l in newTextBib if "@" in l]
        for i, e in enumerate(["empty", "empty2", "Gariazzo:2015rra", "prova"]):
            self.assertIn(e, useful[i])


if __name__ == "__main__":
    unittest.main()



"""Module with the classes and functions
that manage the some dialog windows.

This file is part of the physbiblio package.
"""
import ast
import traceback

from PySide6.QtCore import Qt, Signal
from PySide6.QtGui import QTextCursor
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QGridLayout,
    QLineEdit,
    QPlainTextEdit,
    QProgressBar,
    QPushButton,
    QTableWidgetItem,
    QTextEdit,
    QVBoxLayout,
)

try:
    import physbiblio.gui.resourcesPyside6
    from physbiblio.config import configuration_params, pbConfig
    from physbiblio.database import pBDB
    from physbiblio.errors import pBLogger
    from physbiblio.gui.basicDialogs import (
        askDirName,
        askFileNames,
        askSaveFileName,
        askYesNo,
        infoMessage,
    )
    from physbiblio.gui.catWindows import CatsTreeWindow
    from physbiblio.gui.commonClasses import (
        ObjListWindow,
        PBComboBox,
        PBDDTableWidget,
        PBDialog,
        PBImportedTableModel,
        PBLabel,
        PBLabelRight,
        PBTableView,
        PBTrueFalseCombo,
    )
    from physbiblio.gui.errorManager import pBGUILogger
    from physbiblio.strings.gui import DialogWindowsStrings as dwstr
    from physbiblio.webimport.webInterf import physBiblioWeb
except ImportError:
    print("Could not find physbiblio and its modules!")
    print(traceback.format_exc())


class ConfigEditColumns(PBDialog):
    """Extend `PBDialog` to ask the columns
    that must appear in the main table
    """

    excludeCols = [
        "crossref",
        "bibtex",
        "exp_paper",
        "lecture",
        "phd_thesis",
        "review",
        "proceeding",
        "book",
        "noUpdate",
        "bibdict",
        "abstract",
    ]
    moreCols = [
        "title",
        "author",
        "journal",
        "volume",
        "pages",
        "primaryclass",
        "booktitle",
        "reportnumber",
    ]

    def __init__(self, parent=None, previous=None):
        """Extend `PBDialog.__init__` and create the form structure

        Parameters:
            parent: teh parent widget
            previous: list of columns which must appear
                as selected at the beginning
        """
        super(ConfigEditColumns, self).__init__(parent)
        self.gridlayout = None
        self.items = []
        self.listAll = None
        self.listSel = None
        self.allItems = None
        self.selItems = None
        self.previousSelected = (
            previous if previous is not None else pbConfig.params["bibtexListColumns"]
        )
        self.selected = self.previousSelected
        self.result = False
        self.acceptButton = None
        self.cancelButton = None
        self.initUI()

    def onCancel(self):
        """Reject the output (set self.result to False and close)"""
        self.result = False
        self.close()

    def onOk(self):
        """Accept the output and prepare `self.selected`"""
        self.result = True
        self.selected = []
        for row in range(self.listSel.rowCount()):
            self.selected.append(self.listSel.item(row, 0).text())
        self.close()

    def initUI(self):
        """Initialize the `PBDDTableWidget`s and their content,
        plus the buttons and labels
        """
        self.gridlayout = QGridLayout()
        self.setLayout(self.gridlayout)

        self.items = []
        self.listAll = PBDDTableWidget(self, dwstr.colsAvailable)
        self.listSel = PBDDTableWidget(self, dwstr.colsSelected)
        self.gridlayout.addWidget(PBLabel(dwstr.colsDragAndDrop), 0, 0, 1, 2)
        self.gridlayout.addWidget(self.listAll, 1, 0)
        self.gridlayout.addWidget(self.listSel, 1, 1)

        self.allItems = list(pBDB.descriptions["entries"]) + self.moreCols
        self.selItems = self.previousSelected
        isel = 0
        iall = 0
        for col in self.selItems + [i for i in self.allItems if i not in self.selItems]:
            if col in self.excludeCols:
                continue
            item = QTableWidgetItem(col)
            self.items.append(item)
            if col in self.selItems:
                self.listSel.insertRow(isel)
                self.listSel.setItem(isel, 0, item)
                isel += 1
            else:
                self.listAll.insertRow(iall)
                self.listAll.setItem(iall, 0, item)
                iall += 1

        self.acceptButton = QPushButton(dwstr.ok, self)
        self.acceptButton.clicked.connect(self.onOk)
        self.gridlayout.addWidget(self.acceptButton, 2, 0)

        self.cancelButton = QPushButton(dwstr.cancel, self)
        self.cancelButton.clicked.connect(self.onCancel)
        self.cancelButton.setAutoDefault(True)
        self.gridlayout.addWidget(self.cancelButton, 2, 1)


class ConfigWindow(PBDialog):
    """Create a window for editing the configuration settings"""

    def __init__(self, parent=None):
        """Simple extension of `PBDialog.__init__`"""
        super(ConfigWindow, self).__init__(parent)
        self.textValues = []
        self.result = False
        self.selectedCats = None
        self.selectedExps = None
        self.grid = None
        self.acceptButton = None
        self.cancelButton = None
        self.initUI()

    def onCancel(self):
        """Reject the output (set self.result to False and close)"""
        self.result = False
        self.close()

    def onOk(self):
        """Accept the output (set self.result to True and close)"""
        self.result = True
        self.close()

    def editPDFFolder(self):
        """Open a dialog to select a new folder name
        for the PDF path, and save the result
        in the `ConfigWindow` interface
        """
        ix = pbConfig.paramOrder.index("pdfFolder")
        folder = askDirName(
            dir=self.textValues[ix][1].text(),
            title=dwstr.dirPDFSave,
        )
        if folder.strip() != "":
            self.textValues[ix][1].setText(str(folder))

    def editFile(self, paramkey="logFileName", text=dwstr.logFileName, filter="*.log"):
        """Open a dialog to select a new file name
        for a configuration parameter, and save the result
        in the `ConfigWindow` interface

        Parameters:
            paramkey: the parameter name in the configuration dictionary
            text: description in the file dialog
            filter: filter the folder content in the file dialog
        """
        if paramkey not in pbConfig.paramOrder:
            pBLogger.warning(dwstr.invalidParamkey % paramkey)
            return
        ix = pbConfig.paramOrder.index(paramkey)
        fname = askSaveFileName(
            title=text, dir=self.textValues[ix][1].text(), filter=filter
        )
        if fname.strip() != "":
            self.textValues[ix][1].setText(str(fname))

    def editColumns(self):
        """Open a dialog to select and/or reorder the list of columns
        to show in the entries list, and save the result
        in the `ConfigWindow` interface
        """
        ix = pbConfig.paramOrder.index("bibtexListColumns")
        window = ConfigEditColumns(
            self, ast.literal_eval(self.textValues[ix][1].text().strip())
        )
        window.exec()
        if window.result:
            columns = window.selected
            self.textValues[ix][1].setText(str(columns))

    def editDefCats(self):
        """Open a dialog to select a the default categories
        for the imported entries, and save the result
        in the `ConfigWindow` interface
        """
        ix = pbConfig.paramOrder.index("defaultCategories")
        selectCats = CatsTreeWindow(
            parent=self,
            askCats=True,
            expButton=False,
            previous=ast.literal_eval(self.textValues[ix][1].text().strip()),
        )
        selectCats.exec()
        if selectCats.result == "Ok":
            self.textValues[ix][1].setText(str(self.selectedCats))

    def initUI(self):
        """Create and fill the `QGridLayout`"""
        self.setWindowTitle(dwstr.config)

        grid = QGridLayout()
        self.grid = grid
        grid.setSpacing(1)

        i = 0
        for k in pbConfig.paramOrder:
            i += 1
            val = (
                pbConfig.params[k]
                if isinstance(pbConfig.params[k], str)
                else str(pbConfig.params[k])
            )
            grid.addWidget(
                PBLabel(
                    "%s (<i>%s</i>%s)"
                    % (
                        configuration_params[k].description,
                        k,
                        (" - " + dwstr.globalSett)
                        if configuration_params[k].isGlobal
                        else "",
                    )
                ),
                i - 1,
                0,
                1,
                2,
            )
            if k == "bibtexListColumns":
                self.textValues.append([k, QPushButton(val)])
                self.textValues[-1][1].clicked.connect(self.editColumns)
            elif k == "pdfFolder":
                self.textValues.append([k, QPushButton(val)])
                self.textValues[-1][1].clicked.connect(self.editPDFFolder)
            elif k == "loggingLevel":
                try:
                    self.textValues.append(
                        [
                            k,
                            PBComboBox(
                                self,
                                pbConfig.loggingLevels,
                                pbConfig.loggingLevels[int(val)],
                            ),
                        ]
                    )
                except (IndexError, ValueError):
                    pBGUILogger.warning(dwstr.invalidLoggingLevel)
                    self.textValues.append(
                        [
                            k,
                            PBComboBox(
                                self,
                                pbConfig.loggingLevels,
                                pbConfig.loggingLevels[
                                    int(configuration_params["loggingLevel"].default)
                                ],
                            ),
                        ]
                    )
            elif k == "logFileName":
                self.textValues.append([k, QPushButton(val)])
                self.textValues[-1][1].clicked.connect(self.editFile)
            elif k == "defaultCategories":
                self.textValues.append([k, QPushButton(val)])
                self.textValues[-1][1].clicked.connect(self.editDefCats)
            elif configuration_params[k].special == "boolean":
                self.textValues.append([k, PBTrueFalseCombo(self, val)])
            else:
                self.textValues.append([k, QLineEdit(val)])
            grid.addWidget(self.textValues[i - 1][1], i - 1, 2, 1, 2)

        # OK button
        self.acceptButton = QPushButton(dwstr.ok, self)
        self.acceptButton.clicked.connect(self.onOk)
        # width = self.acceptButton.fontMetrics().boundingRect('OK').width() + 7
        # self.acceptButton.setMaximumWidth(width)
        grid.addWidget(self.acceptButton, i, 0)

        # cancel button
        self.cancelButton = QPushButton(dwstr.cancel, self)
        self.cancelButton.clicked.connect(self.onCancel)
        self.cancelButton.setAutoDefault(True)
        # width = self.cancelButton.fontMetrics().boundingRect('Cancel').width() + 7
        # self.cancelButton.setMaximumWidth(width)
        grid.addWidget(self.cancelButton, i, 1)

        self.setGeometry(100, 100, 1000, 30 * i)
        self.setLayout(grid)


class LogFileContentDialog(PBDialog):
    """Create a window for showing the logFile content"""

    title = dwstr.logFileContent

    def __init__(self, parent=None):
        """Instantiate class and create its widgets

        Parameter:
            parent: the parent widget
        """
        PBDialog.__init__(self, parent)
        self.textEdit = None
        self.closeButton = None
        self.clearButton = None
        self.initUI()

    def clearLog(self):
        """Ask confirmation, then eventually clear
        the content of the log file
        """
        if askYesNo(dwstr.clearLogAsk):
            try:
                open(pbConfig.params["logFileName"], "w").close()
            except IOError:
                pBGUILogger.exception(dwstr.clearLogFailClear)
            else:
                infoMessage(dwstr.clearLogDone)
                self.close()

    def initUI(self):
        """Create window layout and buttons,
        read log file content and print it in the `QPlainTextEdit`
        """
        self.setWindowTitle(self.title)

        grid = QVBoxLayout()
        grid.setSpacing(1)

        grid.addWidget(PBLabel(dwstr.logFileRead % pbConfig.params["logFileName"]))
        try:
            with open(pbConfig.params["logFileName"]) as r:
                text = r.read()
        except IOError:
            text = dwstr.clearLogFailRead
            pBLogger.exception(text)
        self.textEdit = QPlainTextEdit(text)
        self.textEdit.setReadOnly(True)
        grid.addWidget(self.textEdit)

        self.closeButton = QPushButton(dwstr.close, self)
        self.closeButton.setAutoDefault(True)
        self.closeButton.clicked.connect(self.close)
        grid.addWidget(self.closeButton)

        self.clearButton = QPushButton(dwstr.clearLogTitle, self)
        self.clearButton.clicked.connect(self.clearLog)
        grid.addWidget(self.clearButton)

        self.setGeometry(100, 100, 800, 800)
        self.setLayout(grid)


class PrintText(PBDialog):
    """Create a window for printing text of command line output"""

    stopped = Signal()
    pbMax = Signal(float)
    pbVal = Signal(float)

    def __init__(
        self, parent=None, title="", progressBar=True, noStopButton=False, message=None
    ):
        """Constructor.
        Set some properties and create the GUI of the dialog

        Parameters:
            parent: the parent widget
            title: the window title. If an empty string,
                "Redirect print" will be used
            progressBar: True (default) if the widget must have a progress bar
            noStopButton (default False): True if the widget must have
                a "stop" button to stop the iterations
            message: a text to be inserted as a `PBLabel` in the dialog
        """
        super(PrintText, self).__init__(parent)
        self._wantToClose = False
        self.grid = None
        self.progressBar = None
        self.textEdit = None
        self.closeButton = None
        self.cancelButton = None
        if title != "":
            self.title = title
        else:
            self.title = dwstr.redirectPrint
        self.setProgressBar = progressBar
        self.noStopButton = noStopButton
        self.message = message
        self.pbMax.connect(self.progressBarMax)
        self.pbVal.connect(self.progressBarValue)
        self.initUI()

    def keyPressEvent(self, e):
        """Intercept press keys and only exit if escape is pressed
        when closing is enabled

        Parameters:
            e: the `PySide6.QtGui.QKeyEvent`
        """
        if e.key() == Qt.Key_Escape and self._wantToClose:
            self.close()

    def closeEvent(self, event):
        """Manage the `closeEvent` of the dialog.
        Reject unless `self._wantToClose` is True

        Parameter:
            event: a `QEvent`
        """
        if self._wantToClose:
            super(PrintText, self).closeEvent(event)
        else:
            event.ignore()

    def initUI(self):
        """Create the main `QTextEdit`, the buttons and the `QProgressBar`"""
        self.setWindowTitle(self.title)

        grid = QGridLayout()
        self.grid = grid
        grid.setSpacing(1)

        if self.message is not None and self.message.strip() != "":
            grid.addWidget(PBLabel("%s" % self.message))

        self.textEdit = QTextEdit()
        grid.addWidget(self.textEdit)

        if self.setProgressBar:
            self.progressBar = QProgressBar(self)
            grid.addWidget(self.progressBar)

        # cancel button...should learn how to connect it with a thread kill
        if not self.noStopButton:
            self.cancelButton = QPushButton(dwstr.stop, self)
            self.cancelButton.clicked.connect(self.stopExec)
            self.cancelButton.setAutoDefault(True)
            grid.addWidget(self.cancelButton)
        self.closeButton = QPushButton(dwstr.close, self)
        self.closeButton.setDisabled(True)
        grid.addWidget(self.closeButton)

        self.setGeometry(100, 100, 600, 600)
        self.setLayout(grid)
        self.centerWindow()

    def appendText(self, text):
        """Add the given text to the end of the `self.textEdit` content.

        Parameter:
            text: the string to be appended
        """
        self.textEdit.moveCursor(QTextCursor.End)
        self.textEdit.insertPlainText(text)

    def progressBarMin(self, minimum):
        """Set the minimum value for the progress bar

        Parameter:
            minimum (int or float): the value
        """
        if self.setProgressBar:
            self.progressBar.setMinimum(minimum)

    def progressBarMax(self, maximum):
        """Set the maximum value for the progress bar

        Parameter:
            maximum (int or float): the value
        """
        if self.setProgressBar:
            self.progressBar.setMaximum(maximum)

    def progressBarValue(self, value):
        """Set the current value for the progress bar

        Parameter:
            value (int or float): the value
        """
        if self.setProgressBar:
            self.progressBar.setValue(value)

    def stopExec(self):
        """Stop the iterations through the `stopped` Signal
        and disable the `cancelButton`
        """
        self.cancelButton.setDisabled(True)
        self.stopped.emit()

    def enableClose(self):
        """Enable the close button and set `self._wantToClose` to True"""
        self._wantToClose = True
        self.closeButton.setEnabled(True)
        try:
            self.cancelButton.setEnabled(False)
        except Exception:
            pass


class AdvancedImportDialog(PBDialog):
    """create a window for the advanced import"""

    def __init__(self, parent=None):
        """Simple extension of `PBDialog.__init__`"""
        super(AdvancedImportDialog, self).__init__(parent)
        self.result = False
        self.grid = None
        self.searchStr = None
        self.comboMethod = None
        self.acceptButton = None
        self.cancelButton = None
        self.initUI()

    def onCancel(self):
        """Reject the output (set self.result to False and close)"""
        self.result = False
        self.close()

    def onOk(self):
        """Accept the output (set self.result to True and close)"""
        self.result = True
        self.close()

    def initUI(self):
        """Create and fill the `QGridLayout`"""
        self.setWindowTitle(dwstr.advImpTitle)

        grid = QGridLayout()
        self.grid = grid
        grid.setSpacing(1)

        ##search
        grid.addWidget(PBLabel(dwstr.advImpSearch), 0, 0)
        self.searchStr = QLineEdit("")
        grid.addWidget(self.searchStr, 0, 1)

        grid.addWidget(PBLabel(dwstr.advImpMethod), 1, 0)
        self.comboMethod = PBComboBox(
            self,
            ["INSPIRE-HEP", "ADS-NASA", "arXiv", "DOI"],
            current="INSPIRE-HEP",
        )
        grid.addWidget(self.comboMethod, 1, 1)

        # OK button
        self.acceptButton = QPushButton(dwstr.ok, self)
        self.acceptButton.clicked.connect(self.onOk)
        grid.addWidget(self.acceptButton, 2, 0)

        # cancel button
        self.cancelButton = QPushButton(dwstr.cancel, self)
        self.cancelButton.clicked.connect(self.onCancel)
        self.cancelButton.setAutoDefault(True)
        grid.addWidget(self.cancelButton, 2, 1)

        self.setGeometry(100, 100, 400, 100)
        self.setLayout(grid)
        self.searchStr.setFocus()
        self.centerWindow()


class AdvancedImportSelect(ObjListWindow):
    """create a window for the advanced import"""

    def __init__(self, bibs={}, parent=None):
        """Set some properties and call `initUI`

        Parameters:
            bibs: a dictionary containing the imported bibtex entries.
                Each element should be a dictionary containing at least
                a "bibpars" item, a dictionary with at least the keys
                ["ID", "title", "author", "eprint", "doi"],
                and a boolean "exist" item.
            parent: the parent widget
        """
        self.bibs = bibs
        super(AdvancedImportSelect, self).__init__(parent, gridLayout=True)
        self.checkBoxes = []
        self.result = False
        self.askCats = None
        self.tableModel = None
        self.askCats = None
        self.acceptButton = None
        self.cancelButton = None
        self.initUI()

    def onCancel(self):
        """Reject the output (set self.result to False and close)"""
        self.result = False
        self.close()

    def onOk(self):
        """Accept the output (set self.result to True and close)"""
        self.selected = self.tableModel.selectedElements
        self.result = True
        self.close()

    def keyPressEvent(self, e):
        """Intercept press keys and exit if escape is pressed

        Parameters:
            e: the `PySide6.QtGui.QKeyEvent`
        """
        if e.key() == Qt.Key_Escape:
            self.result = False
            self.close()

    def initUI(self):
        """Create and fill the `QGridLayout`"""
        self.setWindowTitle(dwstr.advImpResults)

        self.currLayout.setSpacing(1)

        self.currLayout.addWidget(PBLabel(dwstr.importSelRes))

        headers = ["ID", "title", "author", "eprint", "doi"]
        for k in self.bibs.keys():
            try:
                self.bibs[k]["bibpars"]["eprint"] = self.bibs[k]["bibpars"]["arxiv"]
            except KeyError:
                pass
            try:
                self.bibs[k]["bibpars"]["author"] = self.bibs[k]["bibpars"]["authors"]
            except KeyError:
                pass
            for f in headers:
                try:
                    self.bibs[k]["bibpars"][f] = self.bibs[k]["bibpars"][f].replace(
                        "\n", " "
                    )
                except KeyError:
                    pass
        self.tableModel = PBImportedTableModel(self, self.bibs, headers)
        self.addFilterInput(dwstr.filterEntries, gridPos=(1, 0))
        self.setProxyStuff(0, Qt.AscendingOrder)
        self.finalizeTable(gridPos=(2, 0, 1, 2))

        i = 3
        self.askCats = QCheckBox(dwstr.askCats, self)
        self.askCats.toggle()
        self.currLayout.addWidget(self.askCats, i, 0)

        # OK button
        self.acceptButton = QPushButton(dwstr.ok, self)
        self.acceptButton.clicked.connect(self.onOk)
        self.currLayout.addWidget(self.acceptButton, i + 1, 0)

        # cancel button
        self.cancelButton = QPushButton(dwstr.cancel, self)
        self.cancelButton.clicked.connect(self.onCancel)
        self.cancelButton.setAutoDefault(True)
        self.currLayout.addWidget(self.cancelButton, i + 2, 0)

    def triggeredContextMenuEvent(self, row, col, event):
        """Does nothing"""
        pass

    def handleItemEntered(self, index):
        """Does nothing"""
        pass

    def cellClick(self, index):
        """Does nothing"""
        pass

    def cellDoubleClick(self, index):
        """Does nothing"""
        pass


class DailyArxivDialog(PBDialog):
    """create a window for the advanced import"""

    def __init__(self, parent=None):
        """Simple extension of `PBDialog.__init__`"""
        super(DailyArxivDialog, self).__init__(parent)
        self.result = False
        self.grid = None
        self.comboSub = None
        self.comboCat = None
        self.acceptButton = None
        self.cancelButton = None
        self.initUI()

    def onCancel(self):
        """Reject the output (set self.result to False and close)"""
        self.result = False
        self.close()

    def onOk(self):
        """Accept the output (set self.result to True and close)"""
        self.result = True
        self.close()

    def updateCat(self, category):
        """Replace the current content of the `self.comboSub` combobox
        with the new list of subcategories of the given category

        Parameter:
            category: the name of the category in the
                `physBiblioWeb.webSearch["arxiv"].categories` dictionary
        """
        self.comboSub.clear()
        self.comboSub.addItems(
            ["--"] + physBiblioWeb.webSearch["arxiv"].categories[category]
        )

    def initUI(self):
        """Create and fill the `QGridLayout`"""
        self.setWindowTitle(dwstr.arxDailyTitle)

        self.grid = QGridLayout()
        self.grid.setSpacing(1)

        ##combo boxes
        self.grid.addWidget(PBLabel(dwstr.arxCat), 0, 0)
        self.comboCat = PBComboBox(
            self, [""] + sorted(physBiblioWeb.webSearch["arxiv"].categories.keys())
        )
        self.comboCat.currentTextChanged.connect(self.updateCat)
        self.grid.addWidget(self.comboCat, 0, 1)

        self.grid.addWidget(PBLabel(dwstr.arxSub), 1, 0)
        self.comboSub = PBComboBox(self, [""])
        self.grid.addWidget(self.comboSub, 1, 1)

        # OK button
        self.acceptButton = QPushButton(dwstr.ok, self)
        self.acceptButton.clicked.connect(self.onOk)
        self.grid.addWidget(self.acceptButton, 2, 0)

        # cancel button
        self.cancelButton = QPushButton(dwstr.cancel, self)
        self.cancelButton.clicked.connect(self.onCancel)
        self.cancelButton.setAutoDefault(True)
        self.grid.addWidget(self.cancelButton, 2, 1)

        self.setLayout(self.grid)

        self.setGeometry(100, 100, 400, 100)
        self.centerWindow()


class DailyArxivSelect(AdvancedImportSelect):
    """create a window for the advanced import"""

    def __init__(self, bibs={}, parent=None):
        """Set some properties and call `initUI`

        Parameters:
            bibs: a dictionary containing the imported bibtex entries.
                Each element should be a dictionary containing at least
                a "bibpars" item, a dictionary with at least the keys
                ["ID", "title", "author", "eprint", "doi"],
                and a boolean "exist" item.
            parent: the parent widget
        """
        self.tableModel = None
        self.askCats = None
        self.acceptButton = None
        self.cancelButton = None
        self.abstractArea = None
        super(DailyArxivSelect, self).__init__(bibs, parent)

    def initUI(self):
        """Initialize the widget content, with the buttons and labels"""
        self.setWindowTitle(dwstr.arxResTitle)

        self.currLayout.setSpacing(1)

        self.currLayout.addWidget(PBLabel(dwstr.importSelRes))

        headers = ["eprint", "type", "title", "author", "primaryclass"]
        self.tableModel = PBImportedTableModel(
            self, self.bibs, headers + ["abstract"], idName="eprint"
        )
        self.addFilterInput(dwstr.filterEntries, gridPos=(1, 0))
        self.setProxyStuff(0, Qt.AscendingOrder)

        self.finalizeTable(gridPos=(2, 0, 1, 2))

        i = 3
        self.askCats = QCheckBox(dwstr.askCats, self)
        self.askCats.toggle()
        self.currLayout.addWidget(self.askCats, i, 0)

        # OK button
        self.acceptButton = QPushButton(dwstr.ok, self)
        self.acceptButton.clicked.connect(self.onOk)
        self.currLayout.addWidget(self.acceptButton, i + 1, 0)

        # cancel button
        self.cancelButton = QPushButton(dwstr.cancel, self)
        self.cancelButton.clicked.connect(self.onCancel)
        self.cancelButton.setAutoDefault(True)
        self.currLayout.addWidget(self.cancelButton, i + 2, 0)

        self.abstractArea = QTextEdit(dwstr.abstract, self)
        self.currLayout.addWidget(self.abstractArea, i + 3, 0, 4, 2)

    def cellClick(self, index):
        """Click action

        Parameter:
            index: a `QModelIndex` instance
        """
        if not index.isValid():
            return
        row = index.row()
        try:
            eprint = str(self.proxyModel.sibling(row, 0, index).data())
        except AttributeError:
            pBLogger.debug(dwstr.invalidData, exc_info=True)
            return
        if "already existing" in eprint:
            eprint = eprint.replace(dwstr.alreadyExisting, "")
        if hasattr(self, "abstractFormulas") and self.abstractFormulas is not None:
            a = self.abstractFormulas(
                self.parent(),
                self.bibs[eprint]["bibpars"]["abstract"],
                customEditor=self.abstractArea,
                statusMessages=False,
            )
            a.doText()
        else:
            pBLogger.debug(dwstr.errorAF % eprint)


class ExportForTexDialog(PBDialog):
    """create a window for the "export for tex" function"""

    def __init__(self, parent=None):
        """Simple extension of `PBDialog.__init__`"""
        super(ExportForTexDialog, self).__init__(parent)
        self.numTexFields = 1
        self.bibName = ""
        self.texFileNames = [""]
        self.texNames = []
        self.update = False
        self.remove = False
        self.reorder = False
        self.result = False
        self.texButtons = None
        self.bibButton = None
        self.removeCheck = None
        self.reorderCheck = None
        self.updateCheck = None
        self.addTexButton = None
        self.acceptButton = None
        self.cancelButton = None
        self.grid = QGridLayout()
        self.setLayout(self.grid)
        self.initUI()

    def readForm(self):
        """Read the form content"""
        self.update = self.updateCheck.isChecked()
        self.remove = self.removeCheck.isChecked()
        self.reorder = self.reorderCheck.isChecked()
        self.bibName = self.bibButton.text()
        if self.bibName == dwstr.selFile:
            self.bibName = ""
        self.texFileNames = []
        for button in self.texButtons:
            txt = button.text()
            if "[" in txt:
                try:
                    txt = ast.literal_eval(txt)
                except ValueError:
                    pBLogger.warning(dwstr.invalidText % s)
            self.texFileNames.append(txt)
        tmp = []
        for fn in self.texFileNames:
            if isinstance(fn, list):
                tmp += fn
            else:
                tmp.append(fn)
        self.texNames = [f for f in tmp if f != dwstr.selFile and f != ""]

    def onAddTex(self):
        """Add a new line for the texs"""
        self.numTexFields += 1
        self.readForm()
        for ix in range(self.numTexFields):
            try:
                a = self.texFileNames[ix]
            except IndexError:
                self.texFileNames.append("")
        self.cleanLayout()
        self.initUI()

    def onCancel(self):
        """Reject the output (set self.result to False and close)"""
        self.result = False
        self.close()

    def onOk(self):
        """Accept the output (set self.result to True and close)"""
        self.result = True
        self.readForm()
        self.close()

    def onAskBib(self):
        """Accept the output (set self.result to True and close)"""
        outFName = askSaveFileName(
            self,
            title=dwstr.whereExportBibs,
            filter="Bibtex (*.bib)",
        )
        if outFName != "":
            self.bibButton.setText(outFName)

    def onAskTex(self, ix):
        """Accept the output (set self.result to True and close)"""
        texFile = askFileNames(
            self,
            title=dwstr.whichTexFiles,
            filter="Latex (*.tex)",
        )
        if texFile != "" and texFile != []:
            self.texButtons[ix].setText("%s" % texFile)

    def initUI(self):
        """Create and fill the `QGridLayout`"""
        self.setWindowTitle(dwstr.exportTexFile)
        self.grid.setSpacing(1)

        self.grid.addWidget(PBLabelRight(dwstr.bibName), 0, 0, 1, 2)
        self.bibButton = QPushButton(
            dwstr.selFile if self.bibName == "" else self.bibName, self
        )
        self.bibButton.clicked.connect(self.onAskBib)
        self.grid.addWidget(self.bibButton, 0, 2, 1, 2)

        self.texButtons = []
        for ix in range(self.numTexFields):
            self.grid.addWidget(PBLabelRight(dwstr.texNames), ix + 1, 0, 1, 2)
            self.texButtons.append(
                QPushButton(
                    dwstr.selFile
                    if self.texFileNames[ix] == ""
                    else "%s" % self.texFileNames[ix],
                    self,
                )
            )
            self.texButtons[ix].clicked.connect(lambda s=False, p=ix: self.onAskTex(p))
            self.grid.addWidget(self.texButtons[ix], ix + 1, 2, 1, 2)

        i = 2 + self.numTexFields
        self.addTexButton = QPushButton(dwstr.exportAddTex, self)
        self.addTexButton.clicked.connect(self.onAddTex)
        self.grid.addWidget(self.addTexButton, i - 1, 2)

        self.removeCheck = QCheckBox(dwstr.exportRemove, self)
        if self.remove:
            self.removeCheck.setChecked(True)
        self.grid.addWidget(self.removeCheck, i, 0, 1, 2)
        self.updateCheck = QCheckBox(dwstr.exportUpdate, self)
        if self.update:
            self.updateCheck.setChecked(True)
        self.grid.addWidget(self.updateCheck, i, 2, 1, 2)
        self.reorderCheck = QCheckBox(dwstr.exportReorder, self)
        if self.reorder:
            self.reorderCheck.setChecked(True)
        self.grid.addWidget(self.reorderCheck, i + 1, 0, 1, 4)

        # OK button
        self.acceptButton = QPushButton(dwstr.ok, self)
        self.acceptButton.clicked.connect(self.onOk)
        self.grid.addWidget(self.acceptButton, i + 2, 1)

        # cancel button
        self.cancelButton = QPushButton(dwstr.cancel, self)
        self.cancelButton.clicked.connect(self.onCancel)
        self.cancelButton.setAutoDefault(True)
        self.grid.addWidget(self.cancelButton, i + 2, 2)

        self.setGeometry(100, 100, 400, 25 * (i + 3))
        self.centerWindow()
